/**
 * Split the query string and hash
 * and return [path, search, hash].
 */
const splitSearchAndHash = (url) => {
    const matchResult = /^(.*?)(\?.*?)?(#.*)?$/.exec(url);
    if (!matchResult) {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore until https://github.com/microsoft/TypeScript/issues/45167
        throw new Error('Failed to split search and hash from the input', {
            cause: url,
        });
    }
    const [, path, search, hash] = matchResult;
    return [path, search, hash];
};
export const passThrough = (request, { original }) => original(request);
export default class InterceptRule {
    constructor(input, init = {}) {
        Object.defineProperty(this, "input", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "init", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "headersArr", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "delayDuration", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 0
        });
        Object.defineProperty(this, "restApplyTimes", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 1
        });
        this.input = input;
        this.init = init;
        if (init.headers) {
            this.headersArr.push(...new Headers(init.headers));
        }
        else if (input instanceof Request) {
            this.headersArr.push(...input.headers);
        }
    }
    delay(ms) {
        this.delayDuration += ms;
        return this;
    }
    reply(replier, init) {
        if (replier === undefined || init) {
            this.replier = new Response(replier, init);
        }
        else {
            this.replier = replier;
        }
        return this;
    }
    isActive() {
        return this.restApplyTimes > 0;
    }
    times(n) {
        this.restApplyTimes = n;
        return this;
    }
    once() {
        return this.times(1);
    }
    twice() {
        return this.times(2);
    }
    thrice() {
        return this.times(3);
    }
    persist() {
        return this.times(Infinity);
    }
    test(input, init) {
        // Inactive rule matches no request.
        if (!this.isActive())
            return false;
        const request = init === undefined && input instanceof Request
            ? input
            : new Request(input, init);
        // Required init in later test.
        let requiredInit = this.init;
        // Test required input.
        if (this.input instanceof Request) {
            // For Request, check in later init.
            requiredInit = new Request(this.input, this.init);
        }
        else if (typeof this.input === 'object') {
            // For object like RegExp and URLPattern, test against request URL.
            if (!this.input.test(request.url))
                return false;
        }
        else {
            // For string, separate out search and hash.
            const [path, search, hash] = splitSearchAndHash(request.url);
            // Parse the required based on the request as the required may be relative.
            const [requiredPath, requiredSearch, requiredHash,] = splitSearchAndHash(new URL(this.input, request.url).href);
            // Check parts before the query string.
            if (path !== requiredPath)
                return false;
            // When required, check query string and fragment.
            if ((requiredSearch && search !== requiredSearch)
                || (requiredHash && hash !== requiredHash))
                return false;
        }
        // Test required init.
        // Check required headers.
        if (this.headersArr.some(([key, value]) => request.headers.get(key) !== value)) {
            return false;
        }
        // Check init other than `headers`, `body` and `window`.
        // Specify keys, as destructuring and Object.keys won't include inherited props.
        return [
            'cache',
            'credentials',
            'integrity',
            'keepalive',
            'method',
            'mode',
            'redirect',
            'referrer',
            'referrerPolicy',
            'signal',
        ].every((key) => (!(key in requiredInit) || request[key] === requiredInit[key]));
    }
    async apply(request, fetchers) {
        this.restApplyTimes -= 1;
        const { replier, delayDuration } = this;
        // Throw if no replier provided.
        if (replier === undefined) {
            // eslint-disable-next-line @typescript-eslint/ban-ts-comment
            // @ts-ignore until https://github.com/microsoft/TypeScript/issues/45167
            throw new Error('No reply body or callback configured on this onfetch rule', {
                cause: this,
            });
        }
        // Apply delay.
        if (delayDuration > 0)
            await new Promise((r) => setTimeout(r, delayDuration));
        // Execute callback.
        const bodyInitOrResponse = typeof replier === 'function'
            // Ensure thrown error to be Error.
            ? await Promise.resolve(replier(request, fetchers))
                .catch((err) => {
                if (err instanceof Error)
                    throw err;
                throw new Error(`The reply callback threw an error: ${String(err)}`);
            })
            : await replier;
        // Form the response.
        const response = bodyInitOrResponse instanceof Response
            // Avoid using used response.
            ? bodyInitOrResponse.clone()
            : new Response(bodyInitOrResponse);
        // Ensure URL is defined.
        if (!response.url) {
            Object.defineProperty(response, 'url', {
                // Remove hash as the standard behavior.
                value: request.url.split('#', 1)[0],
            });
        }
        return response;
    }
}
