import toCloneable from './toCloneable';
/**
 * A fetch receiver
 * - For received requests, message to service worker.
 * - For messaged requests, send to the receiver. <- Intercept here.
 */
export default class Channel {
    constructor(workerContainer) {
        Object.defineProperty(this, "port", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "fulfillList", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "beActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
        Object.defineProperty(this, "registerController", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (workerContainer) => {
                const { controller } = workerContainer;
                if (!controller)
                    return;
                const { port1, port2 } = new MessageChannel();
                port1.onmessage = this.onMessage;
                this.port = port1;
                this.fulfillList = [];
                controller.postMessage({ onfetch: true }, [port2]);
                if (this.beActive)
                    port1.postMessage({ status: 'on' });
            }
        });
        /**
         * For received requests, message to the service worker.
         */
        Object.defineProperty(this, "fetch", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (...args) => {
                const { port, fulfillList } = this;
                if (!port || !fulfillList) {
                    throw new Error('Service worker not ready yet');
                }
                port.postMessage({
                    request: await toCloneable(new Request(...args)),
                    index: fulfillList.length,
                });
                return new Promise((resolve, reject) => {
                    fulfillList.push([resolve, reject]);
                });
            }
        });
        Object.defineProperty(this, "onMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (event) => {
                if (!this.beActive || !event.data)
                    return;
                if ('request' in event.data) {
                    // eslint-disable-next-line no-void
                    void this.onRequestMessage(event);
                }
                if ('response' in event.data) {
                    // eslint-disable-next-line no-void
                    void this.onResponseMessage(event);
                }
            }
        });
        /**
         * For messaged requests, send to the receiver.
         */
        Object.defineProperty(this, "onRequestMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (event) => {
                const { data: { request, index } } = event;
                const { port } = this;
                if (!port) {
                    throw new Error('Service worker not ready yet');
                }
                const responseOrError = await this.fetch(request.url, request).catch((err) => err);
                port.postMessage({
                    response: responseOrError instanceof Error
                        ? responseOrError
                        : await toCloneable(responseOrError),
                    index,
                });
            }
        });
        /**
         * For messaged responses, treat as a previous received request's response.
         */
        Object.defineProperty(this, "onResponseMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (event) => {
                const { data: { response, index } } = event;
                const { fulfillList } = this;
                if (!fulfillList) {
                    throw new Error('Service worker not ready yet');
                }
                const fulfill = fulfillList[index];
                if (!fulfill) {
                    throw new Error('Response came for unknown request');
                }
                if (response instanceof Error) {
                    fulfill[1](response);
                }
                else {
                    fulfill[0](new Response(response.body, response));
                }
                fulfillList[index] = null;
            }
        });
        this.registerController(workerContainer);
        workerContainer.addEventListener('controllerchange', () => {
            this.registerController(workerContainer);
        });
    }
    isActive() {
        return this.beActive;
    }
    /**
     * Start receiving worker messages.
     */
    activate() {
        var _a;
        this.beActive = true;
        (_a = this.port) === null || _a === void 0 ? void 0 : _a.postMessage({ status: 'on' });
    }
    /**
     * Stop receiving worker messages.
     */
    restore() {
        var _a;
        this.beActive = false;
        (_a = this.port) === null || _a === void 0 ? void 0 : _a.postMessage({ status: 'off' });
    }
}
